﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Dynamic;
using System.Linq;
using System.Windows;
using System.Reflection;

namespace WPFDynamicProperty
{
    /// <summary>
    /// MainWindow.xaml 的交互逻辑
    /// </summary>
    public partial class MainWindow : Window
    {
        private Person _person;
        private DynamicBindingProxy<Person> _bindingProxy;
        public MainWindow()
        {
            InitializeComponent();

            _person = new Person() { Name = "xiaoli", Age = 23, Birthday = DateTime.Now };

            _bindingProxy = new DynamicBindingProxy<Person>(_person);//动态绑定实体属性

            DataContext = _bindingProxy;
        }

        private void UpdateName(object sender, RoutedEventArgs e)
        {
            ((dynamic)_bindingProxy).Name = newText.Text;
        }

    }

    public class Person
    {
        public string Name { get; set; }
        public int Age { get; set; }
        public DateTime Birthday { get; set; }
    }

    public class DynamicBindingProxy<T> : DynamicObject, INotifyPropertyChanged
    {
        private static readonly Dictionary<string, Dictionary<string, PropertyInfo>> properties =
       new Dictionary<string, Dictionary<string, PropertyInfo>>();//存放T 的属性
        private readonly T _instance;
        private readonly string _typeName;

        public Dictionary<string, object> Changes { get; private set; }//存储变更的属性

        public bool IsDirty { get { return Changes.Count > 0; } }
        //重置脏数据
        public void Reset()
        {
            Changes.Clear();
            NotifyPropertyChanged("IsDirty");
        }
        public DynamicBindingProxy(T instance)
        {
            _instance = instance;
            var type = typeof(T);
            _typeName = type.FullName;
            if (!properties.ContainsKey(_typeName))
                SetProperties(type, _typeName);
            Changes = new Dictionary<string, object>();
        }

        private static void SetProperties(Type type, string typeName)
        {
            var props = type.GetProperties(BindingFlags.Public | BindingFlags.Instance);
            var dict = props.ToDictionary(prop => prop.Name);
            properties.Add(typeName, dict);
        }

        public override bool TryGetMember(GetMemberBinder binder, out object result)
        {
            if (properties[_typeName].ContainsKey(binder.Name))
            {
                result = properties[_typeName][binder.Name].GetValue(_instance, null);
                return true;
            }
            result = null;
            return false;
        }
        public override bool TrySetMember(SetMemberBinder binder, object value)
        {
            if (properties[_typeName].ContainsKey(binder.Name))
            {
                properties[_typeName][binder.Name].SetValue(_instance, value, null);

                Changes[binder.Name] = value;

                NotifyPropertyChanged(binder.Name);
                return true;
            }
            return false;
        }

        private void NotifyPropertyChanged(string name)
        {
            if (PropertyChanged != null)
            {
                PropertyChanged(this, new PropertyChangedEventArgs(name));
            }
        }

        public event PropertyChangedEventHandler PropertyChanged;
    }

}
